from pathlib import Path
from unittest.mock import MagicMock
from unittest.mock import PropertyMock
from unittest.mock import patch

from ezchlog.config import Config


@patch('ezchlog.config.Config.root_dir', new_callable=PropertyMock)
@patch('ezchlog.config.Config.pyproject', new_callable=PropertyMock)
@patch('ezchlog.config.Config.ezchlogconf', new_callable=PropertyMock)
@patch('ezchlog.config.environ', new_callable=dict)
def test_config_init_no_conf(environ, ezchlogconf, pyproject, root_dir) -> None:
    ezchlogconf.return_value = None
    pyproject.return_value = None
    root_path = Path('/some/root/project')
    root_dir.return_value = root_path
    cfg = Config()
    assert cfg.log_file == root_path / 'CHANGELOG.md'
    assert cfg.log_dir == root_path / '_CHANGELOGS'
    assert cfg.category_list == [
        'Security',
        'Fixed',
        'Changed',
        'Added',
        'Remove',
        'Deprecated',
    ]
    assert cfg.category_default == 'Changed'
    assert cfg.default_changelog == """\
# Changelog
All notable changes to this project will be documented in this file.

The format is based on [Keep a Changelog](http://keepachangelog.com/en/1.1.0/)
and this project adheres to [Semantic Versioning](http://semver.org/spec/v2.0.0.html).
"""


@patch('ezchlog.config.Config.root_dir', new_callable=PropertyMock)
@patch('ezchlog.config.Config.pyproject', new_callable=PropertyMock)
@patch('ezchlog.config.Config.ezchlogconf', new_callable=PropertyMock)
@patch('ezchlog.config.environ', new_callable=dict)
def test_config_init_conf_vars(environ, ezchlogconf, pyproject, root_dir) -> None:
    environ.update({
        'EZCHLOG_LOG_FILE': 'changelogs.md',
        'EZCHLOG_LOG_DIR': '.changelogs',
        'EZCHLOG_CATEGORY_LIST': 'add fix chore',
        'EZCHLOG_CATEGORY_DEFAULT': 'chore',
        'EZCHLOG_DEFAULT_CHANGELOG': '# Changelogs\nChanges:\n',
    })
    ezchlogconf.return_value = None
    pyproject.return_value = None
    root_path = Path('/some/root/project')
    root_dir.return_value = root_path
    cfg = Config()
    assert cfg.log_file == root_path / 'changelogs.md'
    assert cfg.log_dir == root_path / '.changelogs'
    assert cfg.category_list == ['add', 'fix', 'chore']
    assert cfg.category_default == 'chore'
    assert cfg.default_changelog == '# Changelogs\nChanges:\n'


@patch('ezchlog.config.Config.root_dir', new_callable=PropertyMock)
@patch('ezchlog.config.Config.pyproject', new_callable=PropertyMock)
@patch('ezchlog.config.Config.ezchlogconf', new_callable=PropertyMock)
@patch('ezchlog.config.Path.cwd')
@patch('ezchlog.config.environ', new_callable=dict)
def test_config_init_conf_ezchlog(environ, cwd, ezchlogconf, pyproject, root_dir, tmp_path) -> None:
    cwd.return_value = tmp_path
    cfg_toml = '''
log_file = "changelogs.md"
log_dir = ".changelogs"
category_list = ["add", "fix", "chore"]
category_default = "chore"
default_changelog = """
# Changelogs
Changes:
"""
'''
    (tmp_path / '.ezchlog.toml').write_text(cfg_toml, encoding='utf-8')
    ezchlogconf.return_value = tmp_path / '.ezchlog.toml'
    pyproject.return_value = None
    root_dir.return_value = tmp_path
    cfg = Config()
    assert cfg.log_file == tmp_path / 'changelogs.md'
    assert cfg.log_dir == tmp_path / '.changelogs'
    assert cfg.category_list == ['add', 'fix', 'chore']
    assert cfg.category_default == 'chore'
    assert cfg.default_changelog == '# Changelogs\nChanges:\n'


@patch('ezchlog.config.Config.root_dir', new_callable=PropertyMock)
@patch('ezchlog.config.Config.pyproject', new_callable=PropertyMock)
@patch('ezchlog.config.Config.ezchlogconf', new_callable=PropertyMock)
@patch('ezchlog.config.Path.cwd')
@patch('ezchlog.config.environ', new_callable=dict)
def test_config_init_conf_pyproject(environ, cwd, ezchlogconf, pyproject, root_dir, tmp_path) -> None:
    cwd.return_value = tmp_path
    cfg_toml = '''
[tool.ezchlog]
log_file = "changelogs.md"
log_dir = ".changelogs"
category_list = ["add", "fix", "chore"]
category_default = "chore"
default_changelog = """
# Changelogs
Changes:
"""
'''
    ezchlogconf.return_value = None
    (tmp_path / 'pyproject.toml').write_text(cfg_toml, encoding='utf-8')
    pyproject.return_value = tmp_path / 'pyproject.toml'
    root_dir.return_value = tmp_path
    cfg = Config()
    assert cfg.log_file == tmp_path / 'changelogs.md'
    assert cfg.log_dir == tmp_path / '.changelogs'
    assert cfg.category_list == ['add', 'fix', 'chore']
    assert cfg.category_default == 'chore'
    assert cfg.default_changelog == '# Changelogs\nChanges:\n'


def test_ezchlogconf_none(tmp_path) -> None:
    cfg = MagicMock(curr_dir=tmp_path)
    ezchlogconf_func = Config.ezchlogconf.func  # type: ignore
    assert ezchlogconf_func(self=cfg) is None


def test_ezchlogconf_path(tmp_path) -> None:
    subdir = tmp_path / 'sub1' / 'sub2' / 'sub3'
    subdir.mkdir(parents=True)
    cfg_path = tmp_path / 'sub1' / '.ezchlog.toml'
    cfg_path.touch()
    cfg = MagicMock(curr_dir=subdir)
    ezchlogconf_func = Config.ezchlogconf.func  # type: ignore
    assert ezchlogconf_func(self=cfg) == cfg_path


def test_pyproject_none(tmp_path) -> None:
    cfg = MagicMock(curr_dir=tmp_path)
    pyproject_func = Config.pyproject.func  # type: ignore
    assert pyproject_func(self=cfg) is None


def test_pyproject_path(tmp_path) -> None:
    subdir = tmp_path / 'sub1' / 'sub2' / 'sub3'
    subdir.mkdir(parents=True)
    cfg_path = tmp_path / 'sub1' / 'pyproject.toml'
    cfg_path.touch()
    cfg = MagicMock(curr_dir=subdir)
    pyproject_func = Config.pyproject.func  # type: ignore
    assert pyproject_func(self=cfg) == cfg_path


def test_git_dir_none(tmp_path) -> None:
    cfg = MagicMock(curr_dir=tmp_path)
    git_dir_func = Config.git_dir.func  # type: ignore
    assert git_dir_func(self=cfg) is None


def test_git_dir_path(tmp_path) -> None:
    subdir = tmp_path / 'sub1' / 'sub2' / 'sub3'
    subdir.mkdir(parents=True)
    git_path = tmp_path / 'sub1' / '.git'
    git_path.mkdir()
    cfg = MagicMock(curr_dir=subdir)
    git_dir_func = Config.git_dir.func  # type: ignore
    assert git_dir_func(self=cfg) == git_path


def test_root_dir_curr_dir(tmp_path) -> None:
    cfg = MagicMock(
        curr_dir=tmp_path,
        ezchlogconf=None,
        pyproject=None,
        git_dir=None,
    )
    root_dir_func = Config.root_dir.func  # type: ignore
    assert root_dir_func(self=cfg) == tmp_path


def test_root_dir_pyproject_dir(tmp_path) -> None:
    cfg = MagicMock(
        curr_dir=tmp_path,
        ezchlogconf=None,
        pyproject=tmp_path / 'project' / 'pyproject.toml',
        git_dir=None,
    )
    root_dir_func = Config.root_dir.func  # type: ignore
    assert root_dir_func(self=cfg) == tmp_path / 'project'


def test_category_class() -> None:
    cfg = MagicMock(category_list=['toto', 'titi'])
    category_class_func = Config.category_class.func  # type: ignore
    category_class = category_class_func(self=cfg)
    assert [c.name for c in category_class] == ['toto', 'titi']


@patch('ezchlog.config.Config.pyproject', new_callable=PropertyMock)
@patch('ezchlog.config.Config.ezchlogconf', new_callable=PropertyMock)
@patch('ezchlog.config.Path.cwd')
@patch('ezchlog.config.environ', new_callable=dict)
def test_iterable(environ, cwd, ezchlogconf, pyproject, tmp_path) -> None:
    environ.update({
        'EZCHLOG_LOG_FILE': 'changelogs.md',
        'EZCHLOG_LOG_DIR': '.changelogs',
        'EZCHLOG_CATEGORY_LIST': 'add fix chore',
        'EZCHLOG_CATEGORY_DEFAULT': 'chore',
        'EZCHLOG_DEFAULT_CHANGELOG': '# Changelogs\nChanges:\n',
    })
    cwd.return_value = tmp_path
    ezchlogconf.return_value = None
    pyproject.return_value = None
    cfg = Config()
    cfg_dict = dict(cfg)
    assert cfg_dict == dict(
        log_dir=str(tmp_path / '.changelogs'),
        log_file=str(tmp_path / 'changelogs.md'),
        categories='add, fix, chore',
        category_default='chore',
        default_changelog='# Changelogs\nChanges:\n',
    )



from argparse import ArgumentParser
from argparse import Namespace
from enum import Enum
from os import environ
from pathlib import Path
from subprocess import run
from sys import stdin
from tempfile import NamedTemporaryFile
from textwrap import indent
from typing import Optional
from typing import Union

from .ezchlog import EzChLog


class Parser:
    def __init__(self, ezchlog: EzChLog) -> None:
        self.log = ezchlog
        self.parser = ArgumentParser()
        self.parser.add_argument(
            '-n',
            '--dry-run',
            action='store_true',
            help="Dry-run the action",
        )
        subparsers = self.parser.add_subparsers(
            required=True,
            dest='action',
            help="Choose an action, you can also have more help using ACTION --help",
        )
        parser_show_config = subparsers.add_parser('showconfig')
        parser_show_config.set_defaults(func=self.show_config_action)
        parser_show_config.add_argument(
            dest='config_key',
            nargs='?',
            metavar='key',
            default='',
            help="Only show this config key value.",
        )
        parser_add = subparsers.add_parser('add')
        parser_add.set_defaults(func=self.add_action)
        parser_add.add_argument(
            'message',
            help="Your message. Use '-' to open an editor instead.",
        )
        Category = self.log.cfg.category_class
        CategoryDef = getattr(Category, self.log.cfg.category_default)
        category_names = [cat.name for cat in list[Enum](Category)]
        parser_add.add_argument(
            'cat',
            nargs='?',
            metavar='type',
            default=CategoryDef,
            type=lambda s: getattr(Category, s) if s else CategoryDef,
            choices=list(Category),
            help=f"Choose one of {', '.join(category_names)}. Defaut to {CategoryDef.name}.",
        )
        parser_add.add_argument(
            'ref',
            nargs='?',
            default='',
            help="Reference for the log. Default is empty.",
        )
        parser_list = subparsers.add_parser('list')
        parser_list.set_defaults(func=self.list_action)
        parser_merge = subparsers.add_parser('merge')
        parser_merge.add_argument(
            'version',
            help="The next version",
        )
        parser_merge.set_defaults(func=self.merge_action)

    def show_config_action(self, opts: Namespace) -> None:
        def format_value(value: Union[str, bool, int, None], *, with_indent: bool = True) -> str:  # noqa: FBT001
            if with_indent and isinstance(value, str) and '\n' in value:
                return '\n' + indent(value, '  ')
            else:
                return f'{value}'

        if opts.config_key:
            value = dict(self.log.cfg).get(opts.config_key)
            print(format_value(value, with_indent=False))  # noqa: T201
        else:
            for key, value in self.log.cfg:
                print(f'{key} = {format_value(value)}')  # noqa: T201

    def open_editor(self, file_ext: str, default_message: str) -> str:
        editor = environ.get('EZCHLOG_EDITOR', environ.get('EDITOR', environ.get('VISUAL', 'vim')))
        if not editor or not stdin.isatty():
            raise Exception(f"Cannot run editor '{editor}'")
        with NamedTemporaryFile(mode='w+', encoding='utf-8', suffix=f'.{file_ext}', delete=False) as f:
            f.write(default_message)
            f.flush()
            try:
                run([editor, f.name])
                f.seek(0)
                return '\n'.join(line for line in f.read().split('\n') if not line.startswith('#'))
            finally:
                Path(f.name).unlink()

    def add_action(self, opts: Namespace) -> None:
        if opts.message == '-':
            opts.message = self.open_editor('md', """
# This a markdown log file.
# Any comment will be removed.
# An empty file will abort.
""")
        message = opts.message.strip()
        if not message:
            raise Exception("Aborted")
        file_path, md_message = self.log.add(dry_run=opts.dry_run, message=message, cat=opts.cat, ref=opts.ref)
        print(file_path)  # noqa: T201
        if opts.dry_run:
            print(md_message)  # noqa: T201

    def list_action(self, opts: Namespace) -> None:
        for p in self.log.list():
            print(p)  # noqa: T201

    def merge_action(self, opts: Namespace) -> None:
        changelog = self.log.merge(dry_run=opts.dry_run, next_version=opts.version)
        if opts.dry_run:
            print(changelog)  # noqa: T201

    def parse(self, args: Optional[list[str]] = None) -> None:
        opts = self.parser.parse_args(args=args)
        opts.func(opts)


def run_cli() -> None:
    Parser(EzChLog()).parse()



#!/bin/env python
from pathlib import Path
from sys import argv

from tomllib import load


def bump(ver: list[int], how: str) -> list[int]:
    if how == 'major':
        return [ver[0] + 1, 0, 0]
    elif how == 'minor':
        return [ver[0], ver[1] + 1, 0]
    else:
        return [ver[0], ver[1], ver[2] + 1]


with Path('pyproject.toml').open('rb') as f:
    cfg = load(f)
version = cfg['project']['version']
next_version = '.'.join(str(x) for x in bump(list(int(x) for x in version.split('.')), argv[1]))
print(next_version)  # noqa: T201



from enum import Enum
from enum import EnumMeta
from pathlib import Path
from unittest.mock import MagicMock
from unittest.mock import patch

from pytest import fixture

from ezchlog.ezchlog import EzChLog


@patch('ezchlog.ezchlog.Config')
def test_ezchlog_init(config_class) -> None:
    cfg = config_class()
    ezchlog = EzChLog()
    assert ezchlog.cfg == cfg


def test_ezchlog_get_slug() -> None:
    assert EzChLog.get_slug('test') == 'test'
    assert EzChLog.get_slug(
        '(a simple)  message w!th accented letter$: « Café », symbols and too long',
    ) == 'a_simple_message_wth_accented_letter_café_symbols'


@patch('ezchlog.ezchlog.Config')
def test_ezchlog_add_simple(config_class, tmp_path) -> None:
    config_class.return_value = MagicMock(
        log_dir=tmp_path / 'logs',
        root_dir=tmp_path,
    )
    ezchlog = EzChLog()
    cat = MagicMock()
    cat.name = 'changed'
    log_file, msg = ezchlog.add(dry_run=False, message="- simple message", cat=cat, ref="")
    assert log_file == Path('logs') / 'changed' / 'simple_message.md'
    assert msg == "- simple message"
    content = (tmp_path / log_file).read_text(encoding='utf-8')
    assert content == "- simple message\n"


@patch('ezchlog.ezchlog.Config')
def test_ezchlog_add_complex(config_class, tmp_path) -> None:
    config_class.return_value = MagicMock(
        log_dir=tmp_path / 'logs',
        root_dir=tmp_path,
    )
    ezchlog = EzChLog()
    cat = MagicMock()
    cat.name = 'changed'
    log_file, msg = ezchlog.add(dry_run=False, message="complex message\nwith\nother\nlines", cat=cat, ref="42")
    assert log_file == Path('logs') / 'changed' / '42_complex_message.md'
    assert msg == "- complex message (42)  \nwith  \nother  \nlines"
    content = (tmp_path / log_file).read_text(encoding='utf-8')
    assert content == "- complex message (42)  \nwith  \nother  \nlines\n"


@fixture
def category_class() -> EnumMeta:
    category_class = Enum('Category', names=['fix', 'add'])  # type: ignore
    return category_class


@patch('ezchlog.ezchlog.Config')
def test_ezchlog_list(config_class, tmp_path, category_class) -> None:
    config_class.return_value = MagicMock(
        log_dir=tmp_path / 'logs',
        category_class=category_class,
    )
    (tmp_path / 'logs' / 'add').mkdir(parents=True)
    (tmp_path / 'logs' / 'add' / 'toto.md').touch()
    (tmp_path / 'logs' / 'add' / 'a_test.md').touch()
    (tmp_path / 'logs' / 'fix').mkdir(parents=True)
    (tmp_path / 'logs' / 'fix' / 'some_fix.md').touch()
    ezchlog = EzChLog()
    assert ezchlog.list() == [
        Path('fix') / 'some_fix.md',
        Path('add') / 'a_test.md',
        Path('add') / 'toto.md',
    ]


@patch('ezchlog.ezchlog.Config')
def test_ezchlog_merge_empty_version(config_class, tmp_path, category_class) -> None:
    log_dir = tmp_path / 'logs'
    log_file = tmp_path / 'chglogs.md'
    default_changelog = '# Changelogs'
    config_class.return_value = MagicMock(
        log_dir=log_dir,
        log_file=log_file,
        category_class=category_class,
        default_changelog=default_changelog,
    )
    ezchlog = EzChLog()
    expected_content = "# Changelogs\n"
    assert ezchlog.merge(dry_run=False, next_version="1.0.0") == expected_content
    assert log_file.read_text(encoding='utf-8') == expected_content


@patch('ezchlog.ezchlog.Config')
def test_ezchlog_merge_not_existing(config_class, tmp_path, category_class) -> None:
    log_dir = tmp_path / 'logs'
    log_file = tmp_path / 'chglogs.md'
    default_changelog = '# Changelogs'
    config_class.return_value = MagicMock(
        log_dir=log_dir,
        log_file=log_file,
        category_class=category_class,
        default_changelog=default_changelog,
    )
    (log_dir / 'add').mkdir(parents=True)
    (log_dir / 'add' / 'toto.md').write_text("- toto content", encoding='utf-8')
    (log_dir / 'add' / 'a_test.md').write_text("- test content", encoding='utf-8')
    (log_dir / 'fix').mkdir(parents=True)
    (log_dir / 'fix' / 'some_fix.md').write_text("- fix content", encoding='utf-8')
    ezchlog = EzChLog()
    expected_content = """\
# Changelogs

## 1.0.0
### fix
- fix content
### add
- test content
- toto content
"""
    assert ezchlog.merge(dry_run=False, next_version="1.0.0") == expected_content
    assert log_file.read_text(encoding='utf-8') == expected_content
    assert not (log_dir / 'add' / 'toto.md').exists()
    assert not (log_dir / 'add' / 'a_test.md').exists()
    assert not (log_dir / 'fix' / 'some_fix.md').exists()


@patch('ezchlog.ezchlog.Config')
def test_ezchlog_merge_existing(config_class, tmp_path, category_class) -> None:
    log_dir = tmp_path / 'logs'
    log_file = tmp_path / 'chglogs.md'
    default_changelog = '# Changelogs'
    config_class.return_value = MagicMock(
        log_dir=log_dir,
        log_file=log_file,
        category_class=category_class,
        default_changelog=default_changelog,
    )
    log_file.write_text("""\
# Changelogs

## 1.0.0
### fix
- fix content
### add
- test content
- toto content
""", encoding='utf-8')
    (log_dir / 'fix').mkdir(parents=True)
    (log_dir / 'fix' / 'minor_modif.md').write_text("- a minor modif", encoding='utf-8')
    ezchlog = EzChLog()
    expected_content = """\
# Changelogs

## 1.1.0
### fix
- a minor modif

## 1.0.0
### fix
- fix content
### add
- test content
- toto content
"""
    assert ezchlog.merge(dry_run=False, next_version="1.1.0") == expected_content
    assert log_file.read_text(encoding='utf-8') == expected_content
    assert not (log_dir / 'fix' / 'minor_modif.md').exists()



from collections.abc import Iterator
from enum import Enum
from enum import EnumMeta
from functools import cached_property
from os import environ
from pathlib import Path
from typing import Any
from typing import Optional
from typing import Union

try:
    from tomllib import load  # type: ignore[import-not-found]
except ModuleNotFoundError:
    from tomli import load  # type: ignore

DEFAULT_MAIN_CHANGELOG = 'CHANGELOG.md'
DEFAULT_CHANGELOG_DIR = '_CHANGELOGS'
DEFAULT_CATEGORIES = [
    'Security',
    'Fixed',
    'Changed',
    'Added',
    'Remove',
    'Deprecated',
]
DEFAULT_DEFAULT_CATEGORY = 'Changed'
DEFAULT_CHANGELOG_TEXT = """\
# Changelog
All notable changes to this project will be documented in this file.

The format is based on [Keep a Changelog](http://keepachangelog.com/en/1.1.0/)
and this project adheres to [Semantic Versioning](http://semver.org/spec/v2.0.0.html).
"""


class Config:
    curr_dir: Path
    log_file: Path
    log_dir: Path
    category_list: list[str]
    category_default: str
    default_changelog: str

    def __init__(self) -> None:
        self.curr_dir = Path.cwd().resolve()
        cfg: dict[str, Any] = {}
        if self.ezchlogconf:
            with self.ezchlogconf.open('rb') as f:
                cfg = load(f)
        elif self.pyproject:
            with self.pyproject.open('rb') as f:
                cfg = load(f).get('tool', {}).get('ezchlog', {})
        self.log_file = Path(environ.get('EZCHLOG_LOG_FILE', cfg.get('log_file', DEFAULT_MAIN_CHANGELOG)))
        if not self.log_file.is_absolute():
            self.log_file = self.root_dir / self.log_file
        self.log_dir = Path(environ.get('EZCHLOG_LOG_DIR', cfg.get('log_dir', DEFAULT_CHANGELOG_DIR)))
        if not self.log_dir.is_absolute():
            self.log_dir = self.root_dir / self.log_dir
        category_list: Union[str, list[str]] = environ.get('EZCHLOG_CATEGORY_LIST', cfg.get('category_list', DEFAULT_CATEGORIES))
        self.category_list = category_list.split(' ') if isinstance(category_list, str) else category_list
        self.category_default = environ.get('EZCHLOG_CATEGORY_DEFAULT', cfg.get('category_default', DEFAULT_DEFAULT_CATEGORY))
        self.default_changelog = environ.get('EZCHLOG_DEFAULT_CHANGELOG', cfg.get('default_changelog', DEFAULT_CHANGELOG_TEXT))

    @cached_property
    def ezchlogconf(self) -> Optional[Path]:
        p_root = self.curr_dir
        for p in [p_root] + list(p_root.parents):
            pf = p / '.ezchlog.toml'
            if pf.is_file():
                return pf
        else:
            return None

    @cached_property
    def pyproject(self) -> Optional[Path]:
        p_root = self.curr_dir
        for p in [p_root] + list(p_root.parents):
            pf = p / 'pyproject.toml'
            if pf.is_file():
                return pf
        else:
            return None

    @cached_property
    def git_dir(self) -> Optional[Path]:
        p_root = self.curr_dir
        for p in [p_root] + list(p_root.parents):
            pd = p / '.git'
            if pd.is_dir():
                return pd
        else:
            return None

    @cached_property
    def root_dir(self) -> Path:
        for attr in ('ezchlogconf', 'pyproject', 'git_dir'):
            if (path := getattr(self, attr)):
                return path.parent
        else:
            return self.curr_dir

    @cached_property
    def category_class(self) -> EnumMeta:
        category_class = Enum('Category', names=self.category_list)  # type: ignore
        return category_class

    def __iter__(self) -> Iterator[tuple[str, Union[str, bool, int]]]:
        Category = self.category_class
        category_names = [cat.name for cat in list[Enum](Category)]
        d: dict[str, Union[str, bool, int]] = dict(
            log_dir=str(self.log_dir),
            log_file=str(self.log_file),
            categories=', '.join(category_names),
            category_default=self.category_default,
            default_changelog=self.default_changelog,
        )
        return iter(d.items())



from argparse import ArgumentParser
from argparse import Namespace
from enum import Enum
from enum import EnumMeta
from pathlib import Path
from unittest.mock import MagicMock
from unittest.mock import patch

from pytest import fixture
from pytest import raises

from ezchlog.cli import Parser
from ezchlog.cli import run_cli


@patch('ezchlog.cli.EzChLog')
@patch('ezchlog.cli.Parser')
def test_run_cli(parser_cls, ezchlog_cls) -> None:
    ezchlog_instance = MagicMock()
    ezchlog_cls.return_value = ezchlog_instance
    parser_instance = MagicMock()
    parser_cls.return_value = parser_instance
    run_cli()
    ezchlog_cls.assert_called_once_with()
    parser_cls.assert_called_once_with(ezchlog_instance)
    parser_instance.parse.assert_called_once_with()


@fixture
def category_class() -> EnumMeta:
    category_class = Enum('Category', names=['toto', 'titi'])  # type: ignore
    return category_class


@fixture
def ezchlog(category_class) -> MagicMock:
    cfg = MagicMock(category_class=category_class, category_default=next(iter(category_class)).name)
    ezchlog = MagicMock(cfg=cfg)
    return ezchlog


def test_parser_init(ezchlog) -> None:
    parser = Parser(ezchlog)
    assert parser
    assert parser.log is ezchlog
    assert isinstance(parser.parser, ArgumentParser)


@patch('ezchlog.cli.Parser.merge_action')
@patch('ezchlog.cli.Parser.list_action')
@patch('ezchlog.cli.Parser.add_action')
@patch('ezchlog.cli.Parser.show_config_action')
def test_parser_parse(show_config_action, add_action, list_action, merge_action, ezchlog, category_class) -> None:
    parser = Parser(ezchlog)
    # showconfig
    parser.parse(args=['-n', 'showconfig'])
    showconfig_ns = Namespace(dry_run=True, action='showconfig', config_key='', func=show_config_action)
    show_config_action.assert_called_once_with(showconfig_ns)
    show_config_action.reset_mock()
    add_action.assert_not_called()
    list_action.assert_not_called()
    merge_action.assert_not_called()
    # add, first case
    parser.parse(args=['add', 'a message', 'titi', '42'])
    add_ns = Namespace(dry_run=False, action='add', message='a message', cat=category_class.titi, ref='42', func=add_action)
    show_config_action.assert_not_called()
    add_action.assert_called_once_with(add_ns)
    add_action.reset_mock()
    list_action.assert_not_called()
    merge_action.assert_not_called()
    # add, second case
    parser.parse(args=['add', '-'])
    add_ns = Namespace(dry_run=False, action='add', message='-', cat=category_class.toto, ref='', func=add_action)
    show_config_action.assert_not_called()
    add_action.assert_called_once_with(add_ns)
    add_action.reset_mock()
    list_action.assert_not_called()
    merge_action.assert_not_called()
    # list
    parser.parse(args=['list'])
    list_ns = Namespace(dry_run=False, action='list', func=list_action)
    show_config_action.assert_not_called()
    add_action.assert_not_called()
    list_action.assert_called_once_with(list_ns)
    list_action.reset_mock()
    merge_action.assert_not_called()
    # merge
    parser.parse(args=['merge', '1.2.3'])
    merge_ns = Namespace(dry_run=False, action='merge', version='1.2.3', func=merge_action)
    show_config_action.assert_not_called()
    add_action.assert_not_called()
    list_action.assert_not_called()
    merge_action.assert_called_once_with(merge_ns)
    merge_action.reset_mock()


def test_parser_showconfig(capsys) -> None:
    parser = MagicMock()
    parser.log.cfg = [
        ('some_key', 'some_value'),
        ('other_key', 42),
    ]
    Parser.show_config_action(self=parser, opts=MagicMock(config_key=''))
    captured = capsys.readouterr()
    assert captured.out == "some_key = some_value\nother_key = 42\n"
    Parser.show_config_action(self=parser, opts=MagicMock(config_key='other_key'))
    captured = capsys.readouterr()
    assert captured.out == "42\n"


def test_parser_add_full(capsys, category_class) -> None:
    parser = MagicMock()
    parser.log.add.return_value = (Path("_CHANGELOGS/titi/a_test_message.md"), "- a test message")
    Parser.add_action(self=parser, opts=MagicMock(dry_run=True, message="a test message", cat=category_class.titi, ref="324"))
    captured = capsys.readouterr()
    assert captured.out == "_CHANGELOGS/titi/a_test_message.md\n- a test message\n"
    parser.open_editor.assert_not_called()
    parser.log.add.assert_called_once_with(dry_run=True, message="a test message", cat=category_class.titi, ref="324")


def test_parser_add_editor(capsys, category_class) -> None:
    parser = MagicMock()
    parser.open_editor.return_value = "  another test message \n "
    parser.log.add.return_value = (Path("_CHANGELOGS/toto/another_test_message.md"), "- another test message")
    Parser.add_action(self=parser, opts=MagicMock(dry_run=False, message="-", cat=category_class.toto, ref=""))
    captured = capsys.readouterr()
    assert captured.out == "_CHANGELOGS/toto/another_test_message.md\n"
    parser.open_editor.assert_called_once_with('md', """
# This a markdown log file.
# Any comment will be removed.
# An empty file will abort.
""")
    parser.log.add.assert_called_once_with(dry_run=False, message="another test message", cat=category_class.toto, ref="")


def test_parser_add_error(capsys, category_class) -> None:
    parser = MagicMock()
    with raises(Exception) as ex:
        Parser.add_action(self=parser, opts=MagicMock(dry_run=False, message=" ", cat=category_class.toto, ref=""))
        assert ex.value == "Aborted"
    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == ""
    parser.open_editor.assert_not_called()
    parser.log.add.assert_not_called()


@patch('ezchlog.cli.run')
@patch('pathlib.Path.unlink')
@patch('ezchlog.cli.NamedTemporaryFile')
@patch('sys.stdin.isatty')
@patch('ezchlog.cli.environ', {})
def test_parser_open_editor_vim(isatty, named_temporary_file, unlink, run_process) -> None:
    isatty.return_value = True
    f = MagicMock(name="f")
    f.name = '/tmp/temp_file.md'
    f.read.return_value = "A message\n# some comment\nfrom editor\n# explanation\n"
    named_temporary_file.return_value.__enter__.return_value = f
    parser = MagicMock()
    msg = Parser.open_editor(self=parser, file_ext='md', default_message="# explanation\n")
    assert msg == "A message\nfrom editor\n"
    isatty.assert_called_once()
    named_temporary_file.assert_called_once()
    _, temp_kwargs = named_temporary_file.call_args_list[-1]
    assert temp_kwargs.get('encoding') == 'utf-8'
    assert temp_kwargs.get('suffix') == '.md'
    f.write.assert_called_once_with("# explanation\n")
    f.seek.assert_called_once_with(0)
    f.flush.assert_called_once()
    f.read.assert_called_once()
    run_process.assert_called_once_with(['vim', '/tmp/temp_file.md'])
    unlink.assert_called_once()


@patch('ezchlog.cli.run')
@patch('pathlib.Path.unlink')
@patch('ezchlog.cli.NamedTemporaryFile')
@patch('sys.stdin.isatty')
@patch('ezchlog.cli.environ', {'EDITOR': 'nano', 'VISUAL': 'geany'})
def test_parser_open_editor_editor(isatty, named_temporary_file, unlink, run_process) -> None:
    isatty.return_value = True
    f = MagicMock(name="f")
    f.name = '/tmp/some_file.md'
    f.read.return_value = "A message\n# some comment\nfrom nano\n# explanation\n"
    named_temporary_file.return_value.__enter__.return_value = f
    parser = MagicMock()
    msg = Parser.open_editor(self=parser, file_ext='md', default_message="# explanation\n")
    assert msg == "A message\nfrom nano\n"
    isatty.assert_called_once()
    named_temporary_file.assert_called_once()
    _, temp_kwargs = named_temporary_file.call_args_list[-1]
    assert temp_kwargs.get('encoding') == 'utf-8'
    assert temp_kwargs.get('suffix') == '.md'
    f.write.assert_called_once_with("# explanation\n")
    f.seek.assert_called_once_with(0)
    f.flush.assert_called_once()
    f.read.assert_called_once()
    run_process.assert_called_once_with(['nano', '/tmp/some_file.md'])
    unlink.assert_called_once()


@patch('ezchlog.cli.run')
@patch('pathlib.Path.unlink')
@patch('ezchlog.cli.NamedTemporaryFile')
@patch('sys.stdin.isatty')
@patch('ezchlog.cli.environ', {'EZCHLOG_EDITOR': 'emacs', 'EDITOR': 'nano', 'VISUAL': 'geany'})
def test_parser_open_editor_ezchlogeditor(isatty, named_temporary_file, unlink, run_process) -> None:
    isatty.return_value = True
    f = MagicMock(name="f")
    f.name = '/tmp/another_file.rst'
    f.read.return_value = "A message\n# some comment\nfrom emacs\n# will no be kept\n"
    named_temporary_file.return_value.__enter__.return_value = f
    parser = MagicMock()
    msg = Parser.open_editor(self=parser, file_ext='rst', default_message="# will no be kept\n")
    assert msg == "A message\nfrom emacs\n"
    isatty.assert_called_once()
    named_temporary_file.assert_called_once()
    _, temp_kwargs = named_temporary_file.call_args_list[-1]
    assert temp_kwargs.get('encoding') == 'utf-8'
    assert temp_kwargs.get('suffix') == '.rst'
    f.write.assert_called_once_with("# will no be kept\n")
    f.seek.assert_called_once_with(0)
    f.flush.assert_called_once()
    f.read.assert_called_once()
    run_process.assert_called_once_with(['emacs', '/tmp/another_file.rst'])
    unlink.assert_called_once()


@patch('ezchlog.cli.run')
@patch('pathlib.Path.unlink')
@patch('ezchlog.cli.NamedTemporaryFile')
@patch('sys.stdin.isatty')
@patch('ezchlog.cli.environ', {'VISUAL': ''})
def test_parser_open_editor_no_editor(isatty, named_temporary_file, unlink, run_process) -> None:
    isatty.return_value = True
    f = MagicMock(name="f")
    f.name = '/tmp/another_file.md'
    f.read.return_value = "A message\n# some comment\nfrom editor\n# will no be kept\n"
    named_temporary_file.return_value.__enter__.return_value = f
    parser = MagicMock()
    with raises(Exception) as ex:
        Parser.open_editor(self=parser, file_ext='md', default_message="# will no be kept\n")
        assert ex.value == "Cannot run editor ''"
    named_temporary_file.assert_not_called()
    f.write.assert_not_called()
    f.seek.assert_not_called()
    f.flush.assert_not_called()
    f.read.assert_not_called()
    run_process.assert_not_called()
    unlink.assert_not_called()


@patch('ezchlog.cli.run')
@patch('pathlib.Path.unlink')
@patch('ezchlog.cli.NamedTemporaryFile')
@patch('sys.stdin.isatty')
@patch('ezchlog.cli.environ', {})
def test_parser_open_editor_not_a_tty(isatty, named_temporary_file, unlink, run_process) -> None:
    isatty.return_value = False
    f = MagicMock(name="f")
    f.name = '/tmp/another_file.md'
    f.read.return_value = "A message\n# some comment\nfrom editor\n# will no be kept\n"
    named_temporary_file.return_value.__enter__.return_value = f
    parser = MagicMock()
    with raises(Exception) as ex:
        Parser.open_editor(self=parser, file_ext='md', default_message="# will no be kept\n")
        assert ex.value == "Cannot run editor 'vim'"
    named_temporary_file.assert_not_called()
    f.write.assert_not_called()
    f.seek.assert_not_called()
    f.flush.assert_not_called()
    f.read.assert_not_called()
    run_process.assert_not_called()
    unlink.assert_not_called()


def test_parser_list(capsys) -> None:
    parser = MagicMock()
    parser.log.list.return_value = [Path("_CHANGELOGS/titi/a_test_message.md"), Path("_CHANGELOGS/toto/another_test_message.md")]
    Parser.list_action(self=parser, opts=MagicMock(dry_run=False))
    captured = capsys.readouterr()
    assert captured.out == "_CHANGELOGS/titi/a_test_message.md\n_CHANGELOGS/toto/another_test_message.md\n"


def test_parser_merge(capsys) -> None:
    parser = MagicMock()
    parser.log.merge.return_value = "The full\nchangelog"
    Parser.merge_action(self=parser, opts=MagicMock(dry_run=True, version="4.2.3"))
    captured = capsys.readouterr()
    assert captured.out == "The full\nchangelog\n"
    parser.log.merge.assert_called_once_with(dry_run=True, next_version="4.2.3")



from enum import Enum
from pathlib import Path
from re import sub

from .config import Config


class StrChainSub:
    def __init__(self, s: str) -> None:
        self.s = s

    def sub(self, regex: str, replacement: str) -> 'StrChainSub':
        return StrChainSub(sub(regex, replacement, self.s))

    def __str__(self) -> str:
        return self.s


class EzChLog:
    def __init__(self) -> None:
        self.cfg = Config()

    @classmethod
    def get_slug(cls, s: str) -> str:
        return str(
            StrChainSub(s.lower())
            .sub(r'\s+', '_')
            .sub(r'\W', '')
            .sub(r'_+', '_'),
        )[:50].strip().strip('_')  # yapf: disable

    def add(self, *, dry_run: bool, message: str, cat: Enum, ref: str) -> tuple[Path, str]:
        first_line = message.split('\n')[0] if '\n' in message else message
        lines = message.split('\n')[1:] if '\n' in message else []
        if not first_line.startswith('- '):
            first_line = '- ' + first_line
        slug = self.get_slug((f'{ref}-' if ref else '') + first_line) + '.md'
        if ref:
            first_line += f" ({ref})"
        lines.insert(0, first_line)
        md_message = '  \n'.join(line.rstrip() for line in lines)
        log_file = self.cfg.log_dir / cat.name / slug
        log_file.parent.mkdir(parents=True, exist_ok=True)
        if not dry_run:
            with log_file.open('w', encoding='utf-8') as f:
                f.write(md_message + '\n')
        return log_file.relative_to(self.cfg.root_dir), md_message

    def list(self) -> list[Path]:
        return [
            p.relative_to(self.cfg.log_dir)
            for cat in list[Enum](self.cfg.category_class)
            for p in sorted((self.cfg.log_dir / cat.name).glob('*.md'))
        ]  # yapf: disable

    def merge(self, *, dry_run: bool, next_version: str) -> str:
        lines_to_insert = [f"## {next_version}"]
        for cat in list[Enum](self.cfg.category_class):
            new_category = True
            for p in sorted((self.cfg.log_dir / cat.name).glob('*.md')):
                if new_category:
                    lines_to_insert.append(f"### {cat.name}")
                    new_category = False
                lines_to_insert.extend(p.read_text(encoding='utf-8').split('\n'))
                p.unlink()
        if self.cfg.log_file.exists():
            lines = self.cfg.log_file.read_text(encoding='utf-8').strip().split('\n')
        else:
            if not dry_run:
                self.cfg.log_file.write_text(self.cfg.default_changelog, encoding='utf-8')
            lines = self.cfg.default_changelog.strip().split('\n')
        if len(lines_to_insert) >= 2:
            pos = 0
            for i, line in enumerate(lines):
                if line.startswith('## '):
                    pos = i
                    break
            else:
                pos = -1
            if pos == -1:
                lines.append("")
                lines.extend(lines_to_insert)
            else:
                lines_to_insert.append("")
                lines = lines[0:pos] + lines_to_insert + lines[pos:]
        changelog = '\n'.join(lines) + '\n'
        if not dry_run:
            self.cfg.log_file.write_text(changelog, encoding='utf-8')
        return changelog
